@@ -14,16 +14,17 @@ use Zonemaster::Backend::Config;
 
 with 'Zonemaster::Backend::DB';
 
+has 'config' => (
+    is       => 'ro',
+    isa      => 'Zonemaster::Backend::Config',
+    required => 1,
+);
+
 has 'dbhandle' => (
     is  => 'rw',
     isa => 'DBI::db',
 );
 
-my $connection_string   = Zonemaster::Backend::Config->load_config()->DB_connection_string( 'mysql' );
-my $connection_args     = { RaiseError => 1, AutoCommit => 1 };
-my $connection_user     = Zonemaster::Backend::Config->load_config()->DB_user();
-my $connection_password = Zonemaster::Backend::Config->load_config()->DB_password();
-
 sub dbh {
     my ( $self ) = @_;
     my $dbh = $self->dbhandle;
@@ -32,6 +33,10 @@ sub dbh {
         return $dbh;
     }
     else {
+        my $connection_string   = $self->config->DB_connection_string( 'mysql' );
+        my $connection_args     = { RaiseError => 1, AutoCommit => 1 };
+        my $connection_user     = $self->config->DB_user();
+        my $connection_password = $self->config->DB_password();
         $dbh = DBI->connect( $connection_string, $connection_user, $connection_password, $connection_args );
         $dbh->{AutoInactiveDestroy} = 1;
         $self->dbhandle( $dbh );
@@ -120,7 +125,7 @@ SELECT id, hash_id FROM test_results WHERE params_deterministic_hash = ? AND (TO
 
         if ( $recent_id ) {
             # A recent entry exists, so return its id
-            if ( $recent_id > Zonemaster::Backend::Config->load_config()->force_hash_id_use_in_API_starting_from_id() ) {
+            if ( $recent_id > $self->config->force_hash_id_use_in_API_starting_from_id() ) {
                 $result_id = $recent_hash_id;
             }
             else {
@@ -145,7 +150,7 @@ SELECT id, hash_id FROM test_results WHERE params_deterministic_hash = ? AND (TO
             my ( $id, $hash_id ) = $dbh->selectrow_array(
                 "SELECT id, hash_id FROM test_results WHERE params_deterministic_hash='$test_params_deterministic_hash' ORDER BY id DESC LIMIT 1" );
                 
-            if ( $id > Zonemaster::Backend::Config->load_config()->force_hash_id_use_in_API_starting_from_id() ) {
+            if ( $id > $self->config->force_hash_id_use_in_API_starting_from_id() ) {
                 $result_id = $hash_id;
             }
             else {
@@ -166,10 +171,10 @@ sub test_progress {
     my $dbh = $self->dbh;
     if ( $progress ) {
         if ($progress == 1) {
-            $dbh->do( "UPDATE test_results SET progress=?, test_start_time=NOW() WHERE $id_field=?", undef, $progress, $test_id );
+            $dbh->do( "UPDATE test_results SET progress=?, test_start_time=NOW() WHERE $id_field=? AND progress <> 100", undef, $progress, $test_id );
         }
         else {
-            $dbh->do( "UPDATE test_results SET progress=? WHERE $id_field=?", undef, $progress, $test_id );
+            $dbh->do( "UPDATE test_results SET progress=? WHERE $id_field=? AND progress <> 100", undef, $progress, $test_id );
         }
     }
 
@@ -200,7 +205,7 @@ sub test_results {
     my $id_field = $self->_get_allowed_id_field_name($test_id);
     
     if ( $new_results ) {
-        $self->dbh->do( qq[UPDATE test_results SET progress=100, test_end_time=NOW(), results = ? WHERE $id_field=?],
+        $self->dbh->do( qq[UPDATE test_results SET progress=100, test_end_time=NOW(), results = ? WHERE $id_field=? AND progress < 100],
             undef, $new_results, $test_id );
     }
 
@@ -219,7 +224,7 @@ sub get_test_history {
     my @results;
     my $sth = {};
 
-    my $use_hash_id_from_id = Zonemaster::Backend::Config->load_config()->force_hash_id_use_in_API_starting_from_id();
+    my $use_hash_id_from_id = $self->config->force_hash_id_use_in_API_starting_from_id();
 
     my $undelegated = "";
     if ($p->{filter} eq "undelegated") {
@@ -335,6 +340,42 @@ sub add_batch_job {
     return $batch_id;
 }
 
+sub build_process_unfinished_tests_select_query {
+    my ( $self ) = @_;
+    
+    if ($self->config->lock_on_queue()) {
+        return "
+            SELECT hash_id, results, nb_retries
+            FROM test_results 
+            WHERE test_start_time < DATE_SUB(NOW(), INTERVAL ".$self->config->MaxZonemasterExecutionTime()." SECOND) 
+            AND nb_retries <= ".$self->config->maximal_number_of_retries()." 
+            AND progress > 0
+            AND progress < 100
+            AND queue = ".$self->config->lock_on_queue();
+    }
+    else {
+        return "
+            SELECT hash_id, results, nb_retries
+            FROM test_results 
+            WHERE test_start_time < DATE_SUB(NOW(), INTERVAL ".$self->config->MaxZonemasterExecutionTime()." SECOND) 
+            AND nb_retries <= ".$self->config->maximal_number_of_retries()." 
+            AND progress > 0
+            AND progress < 100";
+    }
+}
+
+sub process_unfinished_tests_give_up {
+    my ( $self, $result, $hash_id ) = @_;
+
+    $self->dbh->do("UPDATE test_results SET progress = 100, test_end_time = NOW(), results = ? WHERE hash_id=?", undef, encode_json($result), $hash_id);
+}
+
+sub schedule_for_retry {
+    my ( $self, $hash_id ) = @_;
+
+    $self->dbh->do("UPDATE test_results SET nb_retries = nb_retries + 1, progress = 0, test_start_time = NOW() WHERE hash_id=?", undef, $hash_id);
+}
+
 
 no Moose;
 __PACKAGE__->meta()->make_immutable();
